# Add lifespan support for startup/shutdown with strong typing
import os
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from mcp.server.fastmcp import Context, FastMCP
from .logger import get_logger

from .maestro_cli import MaestroCli

logger = get_logger("maestro_mcp")

cli = MaestroCli(
    api_key=os.environ.get("MAESTRO_API_KEY", None),
    maestro_binary_path=os.environ.get("MAESTRO_BINARY_PATH", None),
    api_server=os.environ.get("MAESTRO_API_SERVER", "https://api.copilot.mobile.dev"),
)


@dataclass
class AppContext:
    maestro_cli: MaestroCli


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[AppContext]:
    """Manage application lifecycle with type-safe context"""
    # Initialize on startup
    try:
        logger.info("Starting app")
        yield AppContext(cli)
    finally:
        # Cleanup on shutdown
        pass


# Pass lifespan to server
mcp = FastMCP("Maestro MCP", lifespan=app_lifespan, dependencies=["uv"])


@mcp.resource("https://maestro.dev/cheat-sheet")
def cheat_sheet() -> str:
    """
    A full cheat-sheet of the complete syntax for Maestro flow files.

    :return: a full cheat sheet of Maestro YAML syntax
    """
    try:
        return cli.cheat_sheet()
    except Exception as e:
        logger.error(e.__str__())
        return "Could not retrieve cheat sheet: " + e.__str__()


@mcp.tool()
def get_cheat_sheet(ctx: Context) -> str:
    """
    A full cheat-sheet of the complete syntax for Maestro flow files.

    IMPORTANT: Always call this function to load a summary of Maestro's syntax before using any of the other tools!

    :return: a full cheat sheet of Maestro YAML syntax
    """
    cli: MaestroCli = ctx.request_context.lifespan_context.maestro_cli
    try:
        return cli.cheat_sheet()
    except Exception as e:
        ctx.error(e.__str__())
        raise e


@mcp.tool()
def run_code(ctx: Context, flow_script: str):
    """
    Run a set of Maestro commands (one or more). This can be a full maestro script (including headers), a set of commands (one per line) or simply a single command (eg 'tapOn: 123').

    If this fails due to no device running, please ask the user to start a device!

    *** You don't need to call check_syntax before executing this, as syntax will be checked as part of the execution flow. ***

    :param flow_script: a maestro flow as a string.
    :return:
    """
    cli: MaestroCli = ctx.request_context.lifespan_context.maestro_cli

    try:
        logger.info(f"Executing code:\n\n{flow_script}")
        ctx.info(f"Executing code:\n\n{flow_script}")
        res = cli.run_code(flow_script)
        try:
            curr_hierarchy = cli.get_hierarchy()
            res += f"""

            Current view hierarchy:
            {curr_hierarchy}
            """
        except Exception:
            pass

        return res
    except Exception as e:
        ctx.error(e.__str__())
        raise e


@mcp.tool()
def run_flow_files(ctx: Context, flowFiles: str) -> str:
    """
    Run one or more full Maestro test files.

    If this fails due to no device running, please ask the user to start a device!

    :param flowFiles: a space-separated list of one or more flow files to test. e.g. "flowFile1.yml flowFile2.yml". Please provide the full path to the files.
    :return: result of the flow execution
    """
    cli: MaestroCli = ctx.request_context.lifespan_context.maestro_cli

    try:
        res = cli.run_test(flowFiles)
        try:
            curr_hierarchy = cli.get_hierarchy()
            res += f"""

            Current view hierarchy:
            {curr_hierarchy}
            """
        except Exception:
            pass

        return res
    except Exception as e:
        logger.error(e.__str__())
        ctx.error(e.__str__())
        try:
            curr_hierarchy = cli.get_hierarchy()
            return f"""
            Execution failed with: {e.__str__()}

            Current view hierarchy:
            {curr_hierarchy}
            """
        except Exception:
            pass

        raise e


@mcp.tool()
def start_device(ctx: Context, os: str, platform: str) -> str:
    """
    Start a device for testing.

    DO NOT USE THIS FUNCTION UNLESS THE USER EXPLICITLY REQUESTS IT.

    :param os: the OS of the device. Supported values for iOS: 16, 17 or 18. Supported values for Android: 28, 29, 30, 31, 33
    :param platform: the platform of the device. Supported values: android, ios
    :return:
    """
    cli: MaestroCli = ctx.request_context.lifespan_context.maestro_cli
    logger.info(f"Starting {platform} {os} device...")
    ctx.info(f"Starting {platform} {os} device...")

    try:
        return cli.start_device(os, platform)
    except Exception as e:
        logger.error(e.__str__())
        ctx.error(e.__str__())
        raise e


@mcp.tool()
def get_hierarchy(ctx: Context) -> str:
    """
    Retrieves the view hierarchy of the currently visible screen.

    :return: a JSON of the full view hierarchy of the current screen on the device
    """
    cli: MaestroCli = ctx.request_context.lifespan_context.maestro_cli
    logger.info("Getting view hierarchy...")
    ctx.info("Getting view hierarchy...")

    try:
        return cli.get_hierarchy()
    except Exception as e:
        logger.error(e.__str__())
        ctx.error(e.__str__())
        raise e


@mcp.tool()
def query_docs(ctx: Context, query: str) -> str:
    """
    Queries the Maestro Documentation for instructions related to CLI, Cloud, Maestro Studio or Maestro language Syntax.

    You can include any free text search as well as relevant code blocks in the query.

    :param query: free text search query
    :return: responses according to the official documentation
    """
    cli: MaestroCli = ctx.request_context.lifespan_context.maestro_cli
    logger.info(f"Querying docs for '{query}'...")
    ctx.info(f"Querying docs for '{query}'...")

    try:
        return cli.query_docs("Search the Maestro documentation to answer the following question: " + query)
    except Exception as e:
        logger.error(e.__str__())
        ctx.error(e.__str__())
        raise e


@mcp.tool()
def check_syntax(ctx: Context, code: str) -> str:
    """
    Validates the syntax of a block of Maestro code.
    Valid maestro code must be well-formatted YAML.
    This tool will return an error message if the code is not valid or OK if it is.

    :param code: The maestro code to be validated
    :return: a syntax validation message or OK if code is valid
    """
    cli: MaestroCli = ctx.request_context.lifespan_context.maestro_cli
    logger.info("Checking syntax for %s", code)
    ctx.info("Checking syntax for %s", code)

    try:
        return cli.check_syntax(code)
    except Exception as e:
        logger.error(e.__str__())
        ctx.error(e.__str__())
        raise e


if __name__ == "__main__":
    mcp.run()



import pytest
from maestro_mcp.maestro_cli import MaestroCli


@pytest.fixture
def maestro_cli():
    return MaestroCli(
        api_key=os.environ.get("MAESTRO_API_KEY", None),
        maestro_binary_path=os.environ.get("MAESTRO_BINARY_PATH", None),
        api_server=os.environ.get("MAESTRO_API_SERVER", "https://api.copilot.mobile.dev"),
    )


def test_run_code(maestro_cli):
    res = maestro_cli.run_code("tapOn: login")
    print(res)
    assert res.index("Running on") >= 0


def test_cheat_sheet(maestro_cli):
    res = maestro_cli.cheat_sheet()
    print(res)
    assert "Maestro Flow Script Cheat Sheet" in res


def test_query_docs(maestro_cli):
    res = maestro_cli.query_docs("tapping?")
    print(res)
    assert res.index("tapOn") >= 0


def test_maestro_cli_check_syntax(maestro_cli):
    try:
        maestro_cli.check_syntax("""
        tapO n: "123"
        """)
        assert False # should fail
    except Exception as e:
        print(e)
        assert e.__str__().index("Did you mean `tapOn`?") != -1

    res = maestro_cli.check_syntax("""
    tapOn: "123"
    """)
    print(res)
    assert res == "OK"



import logging
import os
import sys


class Logger:
    """
    A simple logger class that provides the same functionality as the current get_logger.
    """
    def __init__(self, name: str):
        """
        Initialize a logger with the given name.
        
        Args:
            name: The name of the logger
        """
        self.name = name
        self.logger = logging.getLogger(name)
        
        # Configure the logger if it hasn't been configured yet
        if not self.logger.handlers:
            self.logger.setLevel(logging.INFO)
            
            # Create formatter
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            
            # Create file handler only (no console handler)
            log_dir = os.path.expanduser("~/.maestro")
            os.makedirs(log_dir, exist_ok=True)  # Ensure directory exists
            log_file = os.path.join(log_dir, "mcp.log")
            file_handler = logging.FileHandler(log_file)
            file_handler.setLevel(logging.INFO)
            file_handler.setFormatter(formatter)
            
            # Add handler to logger
            self.logger.addHandler(file_handler)
    
    def info(self, message: str, *args, **kwargs):
        """
        Log an info message.
        
        Args:
            message: The message to log
            *args: Arguments for string formatting
            **kwargs: Keyword arguments for string formatting
        """
        if args:
            message = message % args
        self.logger.info(message, **kwargs)
    
    def error(self, message: str, *args, **kwargs):
        """
        Log an error message.
        
        Args:
            message: The message to log
            *args: Arguments for string formatting
            **kwargs: Keyword arguments for string formatting
        """
        if args:
            message = message % args
        self.logger.error(message, **kwargs)


def get_logger(name: str) -> Logger:
    """
    Get a logger instance with the given name.
    This function is provided for backward compatibility.
    
    Args:
        name: The name of the logger
        
    Returns:
        A Logger instance
    """
    return Logger(name)



import os.path
import shutil
import subprocess
import tempfile
import requests
from .logger import get_logger

logger = get_logger("maestro_cli")


class MaestroCli:
    def __init__(
            self,
            api_key: str = None,
            maestro_binary_path: str = None,
            api_server: str = "https://api.copilot.mobile.dev"
    ):
        self.api_server = api_server
        # validate if maestro is in path
        try:
            self.maestro_binary_path = maestro_binary_path or shutil.which("maestro")
            if self.maestro_binary_path is None or not self.maestro_binary_path:
                if os.path.exists(os.path.join(os.path.expanduser("~/.maestro"), "bin", "maestro")):
                    self.maestro_binary_path = os.path.join(os.path.expanduser("~/.maestro"), "maestro")
                    logger.info("maestro binary found at default location %s", self.maestro_binary_path)
                else:
                    logger.info("maestro binary not found in path!")
            else:
                logger.info("maestro binary found at %s", self.maestro_binary_path)
        except Exception as e:
            logger.error("maestro binary not found in path - %s", e)

        self.api_key = api_key
        if api_key is None:
            try:
                home_dir = os.path.expanduser("~")
                with open(os.path.abspath(os.path.join(home_dir, ".mobiledev", "authtoken")), "r") as f:
                    logger.info("maestro api key found at %s",
                                os.path.abspath(os.path.join(home_dir, ".mobiledev", "authtoken")))
                    self.api_key = f.read().strip()
            except Exception as e:
                logger.error(
                    "maestro api key not found, please install it following the instructions at https://docs.maestro.dev/getting-started/installing-maestro and login with maestro login")

    def _respond(self, res) -> str:
        if res.stderr:
            logger.error(res.stderr)
            raise Exception(res.stderr)
        logger.info(res.stdout)
        return res.stdout.strip()

    def _check_cli(self):
        if self.maestro_binary_path is None or not self.maestro_binary_path:
            raise RuntimeError(
                "Maestro CLI not found on path - please install it following the instructions at https://docs.maestro.dev/getting-started/installing-maestro and make sure it's in your PATH, or use the MAESTRO_BINARY_PATH environment variable")

        if self.api_key is None:
            raise RuntimeError(
                "Maestro API key not found - please supply one with the MAESTRO_API_KEY environment variable or use `maestro login` to log in to your maestro account")

    def check_syntax(self, code) -> str:
        self._check_cli()

        with tempfile.NamedTemporaryFile("w") as f:
            f.write(code)
            f.flush()

            # shell out to maestro cli
            res = subprocess.run(
                [self.maestro_binary_path, "check-syntax", f.name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=False,
            )

            return self._respond(res)

    def run_code(self, code) -> str:
        if "---" not in code:  # no header present
            code = "appId: any\n---\n" + code

        lines = code.split("\n")
        header_index = lines.index("---")

        # Keep header as is, process commands after the header
        header = lines[:header_index + 1]
        commands = lines[header_index + 1:]

        # Process each command line
        processed_commands = []
        for line in commands:
            if line.strip() and not line.startswith(" ") and not line.startswith("- "):
                processed_commands.append("- " + line)
            else:
                processed_commands.append(line)

        code = "\n".join(header + processed_commands)

        sc = self.check_syntax(code)
        if sc != 'OK':
            return sc

        with tempfile.NamedTemporaryFile("w") as f:
            f.write(code)
            f.flush()

            # shell out to maestro cli
            res = subprocess.run(
                [self.maestro_binary_path, "test", f.name, "--reinstall-driver=false"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=False,
            )

            return self._respond(res)

    def run_test(self, flow_files):
        self._check_cli()

        res = subprocess.run(
            [self.maestro_binary_path, "test", flow_files, "--reinstall-driver=false"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=False,
        )

        return self._respond(res)

    def get_hierarchy(self) -> str:
        self._check_cli()
        res = subprocess.run(
            [self.maestro_binary_path, "hierarchy", "--reinstall-driver=false", "--compact=true"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=False,
        )

        return self._respond(res)

    def start_device(self, os, platform):
        self._check_cli()

        res = subprocess.run(
            [self.maestro_binary_path, "start-device", "--os-version", os, "--platform", platform],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=False,
        )

        return self._respond(res)

    def cheat_sheet(self) -> str:
        self._check_cli()

        res = requests.get(
            f"{self.api_server}/v2/bot/maestro-cheat-sheet",
            headers={"Authorization": f"Bearer {self.api_key}"},
        )

        if res.status_code != 200:
            raise Exception(f"Failed to get cheat sheet - {res.status_code} {res.text}")

        return res.text

    def query_docs(self, question) -> str:
        self._check_cli()

        # POST https://api.copilot.mobile.dev/v2/bot/query-docs with { question: string }
        question = f"""Search the Maestro documentation to answer the following question:\n{question}\nAlways try to include examples on your responses. No need to include links to documentation."""

        res = requests.post(
            f"{self.api_server}/v2/bot/query-docs",
            headers={"Authorization": f"Bearer {self.api_key}"},
            json={"question": question},
        )
        if res.status_code != 200:
            raise Exception(f"Failed to query docs - {res.status_code} {res.text}")

        return res.json()["answer"]



#!/usr/bin/env python3
"""
Entry point script for running the MCP server during development.
Run with: uv run mcp.py
"""
import sys
import os

# Add the src directory to the path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "src"))

# Now run the module
from maestro_mcp.cli import mcp

if __name__ == "__main__":
    mcp.run()
